from qaravan.core.base_sim import BaseSim
from qaravan.core.utils import string_to_sv, pretty_print_sv
from qaravan.core.circuits import two_local_circ
import numpy as np
from ncon import ncon
from scipy.sparse import csc_matrix
import copy

class StatevectorSim(BaseSim):
    def __init__(self, circ, init_state=None):
        super().__init__(circ, init_state=init_state, nm=None)    

    def initialize_state(self):
        """ 
        internal state is a rank-n tensor with local dimension inherited from the circuit 
        init_state can be provided either as a tensor, a statevector or a bitstring
        """
        if self.init_state is None: 
            sv = np.zeros(self.local_dim**self.num_sites)
            sv[0] = 1.0
            self.state = sv.reshape([self.local_dim]*self.num_sites)
        
        elif type(self.init_state) == np.ndarray:
            if len(self.init_state.shape) == 1:
                self.state = self.init_state.reshape([self.local_dim]*self.num_sites) 
            elif len(self.init_state.shape) == self.num_sites:
                self.state = self.init_state
            else:
                raise ValueError("init_state must be a rank-1 tensor or a rank-n tensor.")

        elif type(self.init_state) == str: 
            sv = string_to_sv(self.init_state, self.circ.local_dim)
            self.state = sv.reshape([self.local_dim]*self.num_sites)

        else:
            raise ValueError("init_state must be either a numpy array or a bitstring.") 
        
    def apply_gate(self, gate):
        self.state = op_action(gate.matrix, gate.indices, self.state, local_dim=self.local_dim)

    def measure(self, meas_sites):
        raise NotImplementedError("Measurement not yet implemented for statevector simulator.")
    
    def local_expectation(self, local_ops):
        """ op is a list of 1-local Hermitian matrices """
        self.run(progress_bar=False)
        right_state = copy.deepcopy(self.state)
        for i in range(self.num_sites):
            state_indices = [-(j+1) for j in range(self.num_sites)] 
            state_indices[i] = 1
            right_state = ncon((local_ops[i], right_state), ([-(i+1),1], state_indices))

        return ncon((self.state.conj(), right_state), ([i for i in range(1, self.num_sites+1)], [i for i in range(1, self.num_sites+1)])).real

    def __str__(self):
        sv = self.state.reshape(self.local_dim**self.num_sites)
        return pretty_print_sv(sv, self.local_dim)

    def get_statevector(self): 
        return self.state.reshape(self.local_dim**self.num_sites)

def locs_to_indices(locs, n): 
    shifted_locs = [loc + 1 for loc in locs]
    gate_indices = [-i for i in shifted_locs] + shifted_locs

    boundaries = [0] + shifted_locs
    tensor_indices = []
    for i in range(len(shifted_locs)):
        tensor_indices += [-j for j in range(boundaries[i] + 1, boundaries[i + 1])]
        tensor_indices.append(shifted_locs[i])
    
    tensor_indices += [-j for j in range(boundaries[-1] + 1, n+1)]
    return gate_indices, tensor_indices

def op_action(op, indices, sv, local_dim=2): 
    if op.ndim != 2*len(indices): 
        op = op.reshape(*[local_dim]*2*len(indices))
    
    n = sv.ndim if sv.ndim > 1 else int(np.log(len(sv)) / np.log(local_dim))
    state = copy.deepcopy(sv).reshape(*[local_dim]*n) if sv.ndim == 1 else copy.deepcopy(sv)
    
    # locs_to_indices assumes ascending order for indices, so sort them first and transpose the operator accordingly
    sorted_indices = sorted(indices)
    sort_order = [indices.index(i) for i in sorted_indices]
    perm = sort_order + [i + len(indices) for i in sort_order]  
    op = op.transpose(perm)

    gate_indices, state_indices = locs_to_indices(sorted_indices, n)
    new_sv = ncon((op, state), (gate_indices, state_indices))
    return new_sv.reshape(local_dim**n) if sv.ndim == 1 else new_sv

def all_zero_sv(num_sites, local_dim=2, dense=False):
    if dense: 
        sv = np.zeros(local_dim**num_sites)
        sv[0] = 1.0
    else: 
        sv = csc_matrix(([1], ([0], [0])), shape=(local_dim**num_sites, 1))
    return sv

def random_sv(num_sites, local_dim=2):
    sv = np.random.rand(local_dim**num_sites) + 1j*np.random.rand(local_dim**num_sites)
    sv /= np.linalg.norm(sv)
    return sv

def partial_overlap(sv1, sv2, local_dim=2, skip=None): 
    system_size = int(np.log(len(sv1)) / np.log(local_dim))
    sites = sorted(skip) if skip is not None else []
    
    psi = sv1.reshape([local_dim] * system_size)
    phi_conj = np.conj(sv2).reshape([local_dim] * system_size)

    psi_labels = [0] * system_size
    phi_conj_labels = [0] * system_size

    next_contract_label = 1
    next_free_label = -1

    for i in range(system_size):
        if i in sites:
            psi_labels[i] = next_free_label
            phi_conj_labels[i] = next_free_label - len(sites)
            next_free_label -= 1
        else:
            psi_labels[i] = next_contract_label
            phi_conj_labels[i] = next_contract_label
            next_contract_label += 1
    
    contraction = ncon([psi, phi_conj], [psi_labels, phi_conj_labels])
    kept_dim = local_dim ** len(sites)
    return contraction.reshape((kept_dim, kept_dim))   

def rdm_from_sv(sv, sites, local_dim=2):
    return partial_overlap(sv, sv, local_dim=local_dim, skip=sites) 

def sv_environment(circ, left_sv, gate_idx): 
    mat, indices = circ[gate_idx].matrix, circ[gate_idx].indices

    circ1 = circ[:gate_idx]
    circ2 = circ[gate_idx+1:]

    sim1 = StatevectorSim(circ1, init_state=None)
    sim2 = StatevectorSim(circ2.dag(), init_state=left_sv)

    sv1 = sim1.run(progress_bar=None).reshape(2**circ.num_sites)
    sv2 = sim2.run(progress_bar=None).reshape(2**circ.num_sites)

    return partial_overlap(sv1, sv2, skip=indices), mat

def environment_state_prep(target_sv, circ=None, skeleton=None, progress_interval=10, max_iter=100, stop_ratio=1e-6):
    """ either circ or skeleton must be provided """
    if circ is None:
        circ = two_local_circ(skeleton)
    
    sim = StatevectorSim(circ)
    ansatz = sim.run(progress_bar=False).reshape(2**circ.num_sites)
    init_cost = np.abs(target_sv.conj().T @ ansatz)
    cost_list = [init_cost]
    for step in range(max_iter):
        for idx in range(len(circ)):
            env, _ = sv_environment(circ, target_sv, idx)
            x,s,yh = np.linalg.svd(env)
            new_mat = yh.conj().T @ x.conj().T
            circ.update_gate(idx, new_mat)
        
        new_cost = 1-np.abs(np.trace(new_mat @ env))
        cost_list.append(new_cost)

        if np.abs(cost_list[-1] - cost_list[-2]) < stop_ratio * cost_list[-2]:
            print(f"Plateau with cost {cost_list[-1]} at step {step}")
            break
        
        if progress_interval is not None and step % progress_interval == 0: 
            print(f"Step {step}: cost = {cost_list[-1]}")

    if step == max_iter - 1: 
        print(f"Max iterations reached with cost {cost_list[-1]}")
        
    return circ, cost_list


import numpy as np

def endian_transform(arr):
    """ only works for qubits now """ 
    if len(arr.shape) == 1: 
        n = int(np.log2(len(arr)))
        new_order = [i for i in range(n)][::-1]
        return arr.reshape(*[2]*n).transpose(*new_order).reshape(2**n)

    elif len(arr.shape) == 2: 
        n = int(np.log2(arr.shape[0]))
        new_order = [i for i in range(n)][::-1]+[i for i in range(n,2*n)][::-1]
        return arr.reshape(*[2]*2*n).transpose(*new_order).reshape(2**n, 2**n)

    else: 
        raise ValueError("only works for 1D and 2D arrays")

def base(b,n,length):
    """ base b representation of number n assuming there are length dits """
    size = int(np.ceil(np.log(max(1,n))/np.log(b)))
    listy = [place
        for i in range(size,-1,-1)
        if (place := n//b**i%b) or i<size] or [0]
    for _ in range(length-len(listy)):
        listy.insert(0, 0)
    return listy

def generate_basis(n): 
    """ generate all possible input strings for n qubits """
    basis = []
    for i in range(2**n):
        basis.append(bin(i)[2:].zfill(n))
    return basis

def string_to_sv(string, local_dim): 
    n = len(string)
    sv = np.zeros(local_dim**n)
    sv[int(string, local_dim)] = 1.0
    return sv

def pretty_print_dm(dmat, local_dim, threshold=1e-3):
    """ prints density matrix as a mixture of quantum states """
    evals, evecs = np.linalg.eigh(dmat)
    n = dmat.shape[0]
    length = int(np.log(n) / np.log(local_dim))

    # Sort eigenvalues and eigenvectors by absolute value of eigenvalues in descending order
    idx = np.argsort(np.abs(evals))[::-1]
    evals = evals[idx]
    evecs = evecs[:, idx]

    result = []
    for i, (eval, evec) in enumerate(zip(evals, evecs.T)):
        if abs(eval) > threshold:
            non_zero_indices = np.where(np.abs(evec) > threshold)[0]
            terms = []
            for idx in non_zero_indices:
                base_repr = ''.join(map(str, base(local_dim, idx, length)))
                term = f"{evec[idx]:.4f}|{base_repr}⟩"
                terms.append(term)
            state_repr = " + ".join(terms)
            result.append(f"{eval.real:.4f} * ({state_repr})")
    
    return "\n ".join(result)
    
def pretty_print_sv(sv, local_dim, threshold=1e-3):
    """ prints statevector as a linear combination of computational basis states """
    n = sv.shape[0]
    length = round(np.log(n) / np.log(local_dim))

    non_zero_indices = np.where(np.abs(sv) > threshold)[0]
    terms = []
    for idx in non_zero_indices:
        base_repr = ''.join(map(str, base(local_dim, idx, length)))
        term = f"{sv[idx]:.4f}|{base_repr}⟩"
        terms.append(term)
    state_repr = " + ".join(terms)
    
    return "".join(state_repr)
    
#################################
##### Fidelities ################
#################################

from scipy.linalg import ishermitian
def sqrtm(hmat): 
    if ishermitian(hmat, atol=1e-8):
        evals, evecs = np.linalg.eigh(hmat)
        evals[evals < 0] = 0.0
        return evecs @ np.diag(np.sqrt(evals.real)) @ evecs.conj().T
    else: 
        raise ValueError("provided matrix is not Hermitian to tolerance 1e-12")

def fidelity(rho1, rho2): 
    if len(rho1.shape) == 1:
        rho1 = np.outer(rho1, rho1.conj())
    if len(rho2.shape) == 1:
        rho2 = np.outer(rho2, rho2.conj())

    rho1_sq = sqrtm(rho1)
    arg = rho1_sq @ rho2 @ rho1_sq
    arg_sq = sqrtm(arg)
    return np.trace(arg_sq).real**2

def hellinger_fidelity(rho1, rho2):
    rho1 = np.asarray(rho1)
    rho2 = np.asarray(rho2)
    if rho1.ndim == 1:
        return np.sum(np.sqrt(rho1 * rho2))**2
    elif rho1.ndim == 2:
        return np.sum(np.sqrt(np.diag(rho1) * np.diag(rho2))).real**2
    else:
        raise ValueError("Input arrays must be either 1D or 2D.")

def two_norm_fidelity(rho1, rho2): 
    if type(rho1) == np.ndarray:
        return np.trace(rho1.conj().T @ rho2).real
    elif isinstance(rho1, MPO): 
        return (rho1.conj() @ rho2).trace(scaled=False).real
    else: 
        raise TypeError(f"provided density operator is not of valid type; must be {np.ndarray} or {MPO} not {type(rho1)}")
    
def vN_entropy(dm): 
    evals = np.linalg.eigvals(dm)
    return sum([-e*np.log(e) for e in evals if e > 0])


from calendar import c
from qaravan.core import BaseSim, pretty_print_sv
from qaravan.tensorQ import MPS, all_zero_mps, string_to_mps, contract_sites, decimate
import numpy as np
from ncon import ncon

class MPSSim(BaseSim):
    def __init__(self, circ, init_state=None):
        super().__init__(circ, init_state=init_state, nm=None)    

    def initialize_state(self):
        """ 
        internal state is an MPS with local dimension inherited from the circuit 
        init_state can be provided either as an MPS or a bitstring (and in the future as a statevector)
        """
        if self.init_state is None: 
            self.state = all_zero_mps(self.num_sites, self.local_dim)

        elif type(self.init_state) == MPS: 
            self.state = self.init_state

        elif type(self.init_state) == str:
            self.state = string_to_mps(self.init_state, self.local_dim) 

        else:
            raise ValueError("init_state must be either an MPS or a bitstring.")
        
    def apply_gate(self, gate):
        inds, mat = gate.indices, gate.matrix
        self.state.canonize(inds)
        sites = self.state.sites
        con_site = contract_sites([sites[i] for i in inds])        
        con_site = ncon((mat, con_site), ([-3,1], [-1,-2,1]))
        dec_sites = decimate(con_site, self.local_dim)
        for i, site in enumerate(dec_sites): 
            sites[inds[i]] = site

    def measure(self, meas_sites):
        raise NotImplementedError("Measurement not yet implemented for MPS simulator.")
    
    def local_expectation(self, local_ops):
        return None
    
    def __str__(self):
        sv = self.state.to_vector()
        return pretty_print_sv(sv, self.local_dim)


from setuptools import setup, find_packages

setup(
    name="qaravan",
    packages=find_packages(where="src"),
    python_requires='>=3.11',
    package_dir={"": "src"},
    version="0.1.12", 
    author="Faisal Alam",
    author_email="mfalam2@illinois.edu",
    description="Unified classical simulation of noiseless and noisy quantum circuits",
    install_requires=[
    "numpy>=2.2.4",
    "scipy>=1.15.2",
    "matplotlib>=3.10.1",
    "tqdm>=4.67.1",
    "ncon>=2.0.0"
    ]
)



import numpy as np
from .gates import *
from .paulis import pauli_commute, pauli_mapping, pauli_multiply, pauli_strings, random_pauli_string

class Noise:
    def get_kraus(self, *args, **kwargs):
        """ returns list of Kraus operators of the noise channel."""
        raise NotImplementedError("Subclasses must implement this method")

    def get_superop(self, *args, **kwargs):
        """ returns the superoperator representation of the noise."""
        kraus_list = self.get_kraus(*args, **kwargs)
        return sum([np.kron(k.conj(), k) for k in kraus_list])

class ThermalNoise(Noise): 
    def __init__(self, t1, t2, one_site_time, two_site_time, local_dim, coupling=1.0):
        self.t1 = t1
        self.t2 = t2
        self.idling_t2 = coupling*t2 + (1-coupling)*t1*2
        self.one_site_time = one_site_time
        self.two_site_time = two_site_time
        self.local_dim = local_dim
        self.coupling = coupling
    
class QubitNoise(ThermalNoise): 
    def __init__(self, t1, t2, one_qubit_time, two_qubit_time, coupling=1.0):
        super().__init__(t1, t2, one_qubit_time, two_qubit_time, 2, coupling)
        
    def get_kraus(self, time, dd=False): 
        t1 = self.t1
        t2 = self.t2 if not dd else self.idling_t2
        p1 = 1-np.exp(-time/t1)
        p2 = 1-np.exp(-time/t2)
        p1c = 1-p1
        p2c = 1-p2
        
        kraus_list = [np.zeros((2,2)) for _ in range(4)]
        
        kraus_list[0][0,0] = np.sqrt(p2)
        kraus_list[0][1,1] = -np.sqrt(p2*p1c)
        
        kraus_list[1][0,1] = -np.sqrt(p1*p2)
        
        kraus_list[2][0,0] = np.sqrt(p2c)
        kraus_list[2][1,1] = np.sqrt(p1c*p2c)
        
        kraus_list[3][0,1] = np.sqrt(p1*p2c)
        
        return kraus_list
        
class QutritNoise(ThermalNoise): 
    def __init__(self, t1, t2, one_qutrit_time, two_qutrit_time, coupling=1.0):
        super().__init__(t1, t2, one_qutrit_time, two_qutrit_time, 3, coupling)
    
    def get_kraus(self, time, dd=False): 
        t1 = self.t1
        t2 = self.t2 if not dd else self.idling_t2
        tr = t1*t2/(t1+t2)
        p1 = 1-np.exp(-time/t1)
        p2 = 1-np.exp(-time/t2)
        pr = 1-np.exp(-time/tr)
        p1c = 1-p1
        p2c = 1-p2
        prc = 1-pr

        kraus_list = [np.zeros((3,3)) for _ in range(9)]

        kraus_list[0][0,0] = np.sqrt(p2c)
        kraus_list[0][1,1] = np.sqrt(prc)
        kraus_list[0][2,2] = np.sqrt(prc)

        kraus_list[1][0,1] = np.sqrt(p2c*p1)
        kraus_list[2][1,2] = np.sqrt(p2c*p1)

        kraus_list[3][0,0] = np.sqrt(p2) * (1/np.sqrt(2))
        kraus_list[3][1,1] = -np.sqrt(p1c*p2) * (1/np.sqrt(2))
        kraus_list[3][2,2] = np.sqrt(p1c*p2) * (1/np.sqrt(2))

        kraus_list[4][0,1] = np.sqrt(p1*p2) * (1/np.sqrt(2))
        kraus_list[5][1,2] = -np.sqrt(p1*p2) * (1/np.sqrt(2))

        kraus_list[6][0,0] = np.sqrt(p2) * (1/np.sqrt(2))
        kraus_list[6][1,1] = np.sqrt(p1c*p2) * (1/np.sqrt(2))
        kraus_list[6][2,2] = -np.sqrt(p1c*p2) * (1/np.sqrt(2))

        kraus_list[7][0,1] = np.sqrt(p1*p2) * (1/np.sqrt(2))
        kraus_list[8][1,2] = np.sqrt(p1*p2) * (1/np.sqrt(2))
        return kraus_list        
    
class PauliNoise(Noise):
    def __init__(self, strings, probs):
        assert np.abs(np.sum(probs) - 1) < 1e-3, f"Probabilities must sum to 1, but got {np.sum(probs)}"
        self.probs = [p for p in probs if np.abs(p) > 1e-6]
        self.strings = [s for s, p in zip(strings, probs) if np.abs(p) > 1e-6]
        self.num_sites = len(strings[0])
        self.local_dim = 2
        
    def sample_string(self):
        return self.strings[np.random.choice(len(self.strings), p=self.probs)]
    
    def symp_matrix(self, full=False): 
        """ symplectic overlap matrix for the terms of the channel."""
        if not full: 
            return np.array([[(-1)**(not pauli_commute(s1, s2)) for s1 in self.strings] for s2 in self.strings])
        else: 
            strings = pauli_strings(self.num_sites)
            return np.array([[(-1)**(not pauli_commute(s1, s2)) for s1 in strings] for s2 in strings])
            
    def ptm(self): 
        strings = pauli_strings(self.num_sites)
        fidelities = []
        for s in strings: 
            fidelity = 0.0
            for s2, p in zip(self.strings, self.probs): 
                if pauli_commute(s, s2): 
                    fidelity += p
                else:
                    fidelity -= p
            fidelities.append(fidelity)
        return np.diag(fidelities)
    
    def fidelities(self):
        """ returns pauli fidelities w.r.t. own strings. """
        return self.symp_matrix() @ self.probs 
    
    def get_kraus(self, *args, **kwargs):
        ops = [embed_operator(len(s), [i for i in range(len(s))], [pauli_mapping[op] for op in s], dense=True) for s in self.strings]
        return [np.emath.sqrt(p)*op for p, op in zip(self.probs, ops)]
        #return [np.sign(p)*np.sqrt(np.abs(p))*op for p, op in zip(self.probs, ops)]
    
    def pinv(self): 
        """ returns the pseudoinverse of the channel. """
        inv_fidelities = 1/np.diag(self.ptm())
        inv_probs = np.linalg.inv(self.symp_matrix(full=True)) @ inv_fidelities
        return PauliNoise(pauli_strings(self.num_sites), inv_probs)
    
    def compose(self, other):
        """ compose two noise channels. """
        new_strings = []
        new_probs = []

        for s1, p1 in zip(self.strings, self.probs):
            for s2, p2 in zip(other.strings, other.probs):
                new_string, phase = pauli_multiply(s1, s2)
                new_strings.append(new_string)
                new_probs.append(p1 * p2)

        unique_strings = list(set(new_strings))
        new_probs_dict = {s: 0 for s in unique_strings}
        for s, p in zip(new_strings, new_probs):
            new_probs_dict[s] += p
        new_probs = [new_probs_dict[s] for s in unique_strings]

        return PauliNoise(unique_strings, new_probs)
    
    def __str__(self): 
        noise_dict = {str(s): np.around(p, 3) for s, p in zip(self.strings, self.probs)}
        return str(noise_dict)

def gate_time(gate, nm):   
    if gate.name[0:4] == "CNOT": 
        return nm.two_site_time
    elif gate.name[0:2] in ["RZ", "CU", "SW"]:
        return 0.0
    else: 
        return nm.one_site_time
    
def random_pauli_channel(n,n_strings=5,c=0.1):
    strings = ['i'*n] + [random_pauli_string(n) for _ in range(n_strings-1)]
    probs = [1,] + [c*np.random.rand() for _ in range(n_strings-1)] 
    probs /= np.sum(probs)
    return PauliNoise(strings, probs)


import numpy as np
import copy 
from scipy.linalg import block_diag, expm 
import scipy.sparse as sp
from functools import reduce
from .paulis import pauli_X, pauli_Y, pauli_Z, pauli_mapping

def embed_operator(num_sites, active_sites, local_ops, local_dim=2, dense=False, factor=False):
    # only works if Pauli string is provided 
    if type(local_ops) == str: 
        local_ops = [pauli_mapping[op] for op in local_ops]

    # only works if active_sites is contiguous
    if type(local_ops) == np.ndarray:  
        full_op = [np.eye(local_dim, dtype=complex) for i in range(num_sites) if i not in active_sites]
        full_op.insert(active_sites[0], local_ops)
        return reduce(np.kron, full_op)
        
    # if local_ops is a list of 1-qubit matrices 
    if dense: 
        full_op = [np.eye(local_dim, dtype=complex) for _ in range(num_sites)]
        for site, op in zip(active_sites, local_ops):
            full_op[site] = op if not factor else op/2
        return reduce(np.kron, full_op)
    
    else: 
        full_op = [sp.eye(local_dim, format='csr', dtype=complex) for _ in range(num_sites)]
        for site, op in zip(active_sites, local_ops):
            full_op[site] = sp.csr_matrix(op) if not factor else sp.csr_matrix(op/2)
        return reduce(sp.kron, full_op)

class Gate:
    def __init__(self, name, indices, matrix, nm=None):
        self.name = name
        self.indices = [indices] if type(indices) == int else indices
        self.span = len(self.indices)
        self.matrix = matrix
        self.nm = nm

    def __str__(self):
        return f"{self.name} gate on site(s) {self.indices}"

    def to_superop(self): 
        if self.nm is None: 
            return np.kron(self.matrix.conj(), self.matrix).reshape([self.get_local_dim()]*4*self.span)
        else: 
            return self.nm.get_superop(self.time, self.dd).reshape([self.get_local_dim()]*4*self.span)
        
    def get_local_dim(self):
        return int(self.matrix.shape[0]**(1/self.span))
    
    def dag(self): 
        g = copy.deepcopy(self)
        g.matrix = g.matrix.conj().T
        return g    
    
class SuperOp:
    """ deprecated; need to be debugged first """
    def __init__(self, name, span, matrix, start_idx=0, time=0.0):
        self.name = name
        
        self.start_idx = start_idx
        self.indices = [start_idx+i for i in range(span)]
        self.span = span
        
        self.matrix = matrix
        self.shape = matrix.shape 
        self.time = time 
        
    def shift(self, new_start_idx): 
        return SuperOp(self.name, self.span, self.matrix, start_idx=new_start_idx, time=self.time)
    
    def __str__(self):
        return f"{self.name} superoperator on site(s) {self.indices}"

###########################################
############# SUBCLASSES ##################
###########################################

class ID(Gate): 
    def __init__(self, local_dim, time, indices, nm=None, dd=False): 
        super().__init__("ID", indices, matrix=None)
        self.matrix = np.eye(local_dim**self.span)
        self.time = time
        self.dd = dd
        self.nm = nm
        
    def __str__(self):
        str = f"Idling on site(s) {self.indices} with {type(self.nm).__name__} for time {self.time}"
        if self.dd: 
            str += "; dynamical decoupling"
        return str
    
# n is needed for PBC where *non-local* CUs are allowed across the boundary
# indices = (targets, control)
# note target indices HAVE to be in ascending order and nearest neighbors
# applies subgate when control wire is at 1 and applies identity otherwise

class CUGate(Gate):
    def __init__(self, indices, subgate, n):
        d = subgate.get_local_dim()
        submat = subgate.matrix
        s = submat.shape[0]
        indices = np.array([int(i) for i in indices])
        r = np.min(np.abs(indices[:-1] - indices[-1]))
        
        if r > 1: # 'long'-range gate across boundary
            raise NotImplementedError("Long range CUGate has not been implemented yet")
            
        else: # nearest neighbor gate 
            mat_list = [np.eye(s), submat] + [np.eye(s)] * (d-2)
            mat = block_diag(*mat_list)
            
            if indices[-2] < indices[-1]: 
                indices = indices
                mat = mat.reshape(d,s,d,s).transpose(1,0,3,2).reshape(d*s,d*s)
                bottom_heavy = False
            elif indices[-1] < indices[0]: 
                indices = np.concatenate(([indices[-1]], indices[0:-1]))
                bottom_heavy = True
            else: 
                raise ValueError("indices does not have allowed ordering")
        
        name = f"CU{d}"
        super().__init__(name, indices, mat)
        self.bottom_heavy = bottom_heavy
        
    def __str__(self):
        title = "bottom heavy" if self.bottom_heavy else "top heavy"
        return f"{title} {self.name} gate on site(s) {self.indices}"
    
# n is needed for PBC where *non-local* CNOTs are allowed across the boundary
# indices = (target, control)
class CNOTGate(Gate):
    """ should become subclass of CUGate """
    def __init__(self, indices, matrix, n):
        d = int(np.sqrt(matrix.shape[0]))
        indices = [int(i) for i in indices]
        
        if np.abs(indices[0] - indices[1]) > 1: # 'long'-range gate across boundary
            temp = np.array(indices) + 1
            if temp[0]%n < temp[1]%n: 
                indices = indices
                mat = matrix
                bottom_heavy = False
            else: 
                indices = indices[::-1]
                mat = matrix.reshape(d,d,d,d).transpose(1,0,3,2).reshape(d*d,d*d)
                bottom_heavy = True
        else: # nearest neighbor gate 
            if indices[0] < indices[1]: 
                indices = indices
                mat = matrix
                bottom_heavy = False
            else: 
                indices = indices[::-1]
                mat = matrix.reshape(d,d,d,d).transpose(1,0,3,2).reshape(d*d,d*d)
                bottom_heavy = True
        
        name = f"CNOT{d}"
        super().__init__(name, indices, mat)
        self.bottom_heavy = bottom_heavy
        
    def __str__(self):
        title = "bottom heavy" if self.bottom_heavy else "top heavy"
        return f"{title} {self.name} gate on site(s) {self.indices}"
    
class ParamGate(Gate):
    def __init__(self, name, indices, *args): 
        super().__init__(name, indices, None) 
        
        if type(args[0]) == np.ndarray: 
            mat = args[0]
            self.update_matrix(mat)
            self.angles = self.solve_angles()
        
        else:
            self.angles = [args] if type(args) == float else args
            mat = self.construct_matrix()
            self.update_matrix(mat)
            
    def __str__(self):
        return f"{self.name} gate on site(s) {self.indices} with angle(s) {self.angles}"

    def update_matrix(self, mat): 
        self.matrix = mat
        
    def construct_matrix(self): 
        raise NotImplementedError("Subclasses must implement this method")
        
    def solve_angles(self): 
        return "unsolved" # if subclass doesn't bother implementing we probably don't need angles
    
########################################
############ QUBIT GATES ###############
########################################

def SX(indices): return Gate("SX", 
                              indices,
                              (1/np.sqrt(2)) * np.array([[1,-1j],[-1j,1]]))

def X(indices): return Gate("X", 
                            indices, 
                            np.array([[0,1],[1,0]]))

def H(indices): return Gate("H", 
                            indices, 
                            np.array([[1,1],[1,-1]])/np.sqrt(2))

def SDG(indices): return Gate("SDG", 
                              indices, 
                              np.array([[1,0],[0,-1j]]))

def CNOT(indices, n): return CNOTGate(indices, 
                                      np.array([[1,0,0,0],[0,0,0,1],[0,0,1,0],[0,1,0,0]]), 
                                      n)  

def CZ(indices): return Gate("CZ", 
                             indices, 
                             np.array([[1,0,0,0], [0,1,0,0], [0,0,1,0], [0,0,0,-1]]))

def SWAP(indices): return Gate("SWAP", indices, 
                                np.array([[1,0,0,0],[0,0,1,0],[0,1,0,0],[0,0,0,1]]))

def iSWAP(indices): return Gate("iSWAP", indices,
                                np.array([[1,0,0,0],[0,0,1j,0],[0,1j,0,0],[0,0,0,1]]))

class RZ(ParamGate): 
    def __init__(self, indices, *args):
        super().__init__("RZ", indices, *args)
    
    def construct_matrix(self): 
        return np.array([[1,0],[0,np.exp(1.j*self.angles[0])]])
    
class RZZ(ParamGate):
    def __init__(self, indices, *args):
        super().__init__("RZZ", indices, *args)
    
    def construct_matrix(self): 
        return expm(-1.j*self.angles[0] * embed_operator(2, [0,1], [pauli_Z, pauli_Z], dense=True))
    
class CPhase(ParamGate):
    def __init__(self, indices, *args):
        super().__init__("CPhase", indices, *args)
    
    def construct_matrix(self): 
        return np.array([[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,np.exp(1.j*self.angles[0])]])
    
    def solve_angles(self): 
        theta = np.angle(self.matrix[3,3])
        return (theta,)

class U(ParamGate):
    def __init__(self, indices, *args):
        super().__init__("U", indices, *args)
    
    def construct_matrix(self): 
        t,p,l = self.angles
        return np.array([[np.cos(t/2), -np.exp(1j*l)*np.sin(t/2)],
                           [np.exp(1j*p)*np.sin(t/2), np.exp(1j*(p+l))*np.cos(t/2)]])
    
    def solve_angles(self): 
        phase = np.angle(self.matrix[0,0]) 
        self.matrix = self.matrix / np.exp(1.j*phase) # removes phase for top left entry
        t = np.arccos(self.matrix[0,0])*2
        if np.abs(np.sin(t/2)) > 1e-12:
            l = np.angle(-self.matrix[0,1]/np.sin(t/2))
            p = np.angle(self.matrix[1,1]/self.matrix[0,0])-l
        else: 
            l_plus_p = np.angle(self.matrix[1,1])
            l,p = l_plus_p/2, l_plus_p/2
        
        return t,p,l        
        
    def decompose(self, basis='ZSX'): 
        if basis == 'ZSX': 
            t,p,l = self.solve_angles()#self.angles
            i = self.indices
            
            return list(reversed([
                RZ(i,p+np.pi),
                SX(i),
                RZ(i,t+np.pi),
                SX(i), 
                RZ(i,l)
            ]))   # If A = BC, we must apply C first and then B
            
        else: 
            raise NotImplementedError(f"{basis} basis decomposition has not been implemented")

class Givens(ParamGate):
    def __init__(self, indices, *args):
        super().__init__("Givens", indices, *args)
    
    def construct_matrix(self): 
        theta = self.angles[0]
        return np.array([[1,0,0,0],
                     [0, np.cos(theta/2), -np.sin(theta/2), 0], 
                     [0, np.sin(theta/2), np.cos(theta/2), 0], 
                     [0,0,0,1]])
     
    def solve_angles(self): 
        theta = np.arccos(self.matrix[1,1])*2
        return (theta,)

class XYCoupling(ParamGate):
    def __init__(self, indices, *args):
        super().__init__("XYCoupling", indices, *args)
    
    def construct_matrix(self): 
        theta = self.angles[0]
        xx = embed_operator(2, [0,1], [pauli_X, pauli_X], dense=True)
        yy = embed_operator(2, [0,1], [pauli_Y, pauli_Y], dense=True)
        return expm(-1.j*theta * (xx + yy))

#########################################
############# QUTRIT GATES ##############
#########################################

def X01(indices): return Gate("X01", 
                              indices, 
                              np.array([[0,-1j,0],[-1j,0,0],[0,0,1]]))

def SX01(indices): return Gate("SX01", 
                               indices, 
                               (1/np.sqrt(2)) * np.array([[1,-1j,0],[-1j,1,0],[0,0,np.sqrt(2)]]))

def X12(indices): return Gate("X12", 
                              indices, 
                              np.array([[1,0,0],[0,0,-1j],[0,-1j,0]]))

def SX12(indices): return Gate("SX12", 
                               indices, 
                               (1/np.sqrt(2)) * np.array([[np.sqrt(2),0,0],[0,1,-1j],[0,-1j,1]]))

def H01(indices): return Gate("H01", 
                              indices, 
                              np.array([[1/np.sqrt(2),1/np.sqrt(2),0],[1/np.sqrt(2),-1/np.sqrt(2),0],[0,0,1]]))

def SDG01(indices): return Gate("SDG01", 
                              indices, 
                              np.array([[1,0,0],[0,-1j,0],[0,0,1]]))

def CNOT3(indices, n): return CNOTGate(indices, 
                                      np.array([[1., 0., 0., 0., 0., 0., 0., 0., 0.],
                                       [0., 0., 0., 0., 1, 0., 0., 0., 0.],
                                       [0., 0., 1/np.sqrt(2), 0., 0., 1/np.sqrt(2), 0., 0., 0.],
                                       [0., 0., 0., 1., 0., 0., 0., 0., 0.],
                                       [0., 1., 0., 0., 0., 0., 0., 0., 0.],
                                       [0., 0.,  -1/np.sqrt(2), 0., 0., 1/np.sqrt(2), 0., 0., 0.],
                                       [0., 0., 0., 0., 0., 0., 1., 0., 0.],
                                       [0., 0., 0., 0., 0., 0., 0., 1j, 0.],
                                       [0., 0., 0., 0., 0., 0., 0., 0., 1.]]).T, 
                                      n)   

def SWAP3(indices): return Gate("SWAP3", indices, 
                                np.array([[1., 0., 0., 0., 0., 0., 0., 0., 0.],
                                       [0., 0., 0., 1., 0., 0., 0., 0., 0.],
                                       [0., 0., 0, 0., 0., 0, 1., 0., 0.],
                                       [0., 1., 0., 0., 0., 0., 0., 0., 0.],
                                       [0., 0., 0., 0., 1., 0., 0., 0., 0.],
                                       [0., 0.,  0, 0., 0., 0, 0., 1., 0.],
                                       [0., 0., 1., 0., 0., 0., 0., 0., 0.],
                                       [0., 0., 0., 0., 0., 1., 0., 0, 0.],
                                       [0., 0., 0., 0., 0., 0., 0., 0., 1.]]))

class RZ01(ParamGate): 
    def __init__(self, indices, *args):
        super().__init__("RZ01", indices, *args)
    
    def construct_matrix(self): 
        return np.array([[1,0,0],[0,np.exp(1.j*self.angles[0]),0],[0,0,1]])
    
class RZ12(ParamGate): 
    def __init__(self, indices, *args):
        super().__init__("RZ12", indices, *args)
    
    def construct_matrix(self): 
        return np.array([[1,0,0],[0,1,0],[0,0,np.exp(1.j*self.angles[0])]])

class U01(ParamGate):
    def __init__(self, indices, *args):
        super().__init__("U01", indices, *args)
    
    def construct_matrix(self): 
        t,p,l = self.angles
        return np.array([[np.cos(t/2), -np.exp(1j*l)*np.sin(t/2), 0],
                           [np.exp(1j*p)*np.sin(t/2), np.exp(1j*(p+l))*np.cos(t/2), 0],
                           [0, 0, 1]])
    
    def solve_angles(self): 
        return None # TODO implement solver 
    
    def decompose(self, basis='ZSX'): 
        if basis == 'ZSX': 
            t,p,l = self.angles
            i = self.indices
            return list(reversed([
                RZ01(i,p+np.pi),
                SX01(i),
                RZ01(i,t+np.pi),
                SX01(i), 
                RZ01(i,l)
            ]))
        else: 
            raise NotImplementedError(f"{basis} basis decomposition has not been implemented")

def random_unitary(size):
    a = np.random.rand(size, size) + 1.j * np.random.rand(size, size)
    h = a @ a.conj().T
    _, u = np.linalg.eigh(h)
    return u

def kak_unitary(params): 
    left_mat = np.kron(U(0, *params[0:3]).matrix, U(0, *params[3:6]).matrix)
    right_mat = np.kron(U(0, *params[9:12]).matrix, U(0, *params[12:15]).matrix)
    
    xx, yy, zz = np.kron(pauli_X, pauli_X), np.kron(pauli_Y, pauli_Y), np.kron(pauli_Z, pauli_Z)
    arg = sum([p*P for p,P in zip(params[6:], [xx, yy, zz])])
    center_mat = expm(1.j* arg)
    
    return left_mat @ center_mat @ right_mat

def is_unitary(u):
    return np.allclose(u @ u.conj().T, np.eye(u.shape[0]), atol=1e-10) and np.allclose(u.conj().T @ u, np.eye(u.shape[0]), atol=1e-10)